from pathlib import Path
from pickle import load
from netfilterqueue import NetfilterQueue
from socket import fromfd, AF_UNIX, SOCK_STREAM
from scapy.all import *
from binascii import b2a_hex
from getmac import get_mac_address
from subprocess import check_output
from time import time
from update_models import get_packet_feature_list
from sys import argv

'''
sudo python3 secure_filtering.py F4:B8:5E:5F:A9:57,live_picture,accept,live_stream,accept
'''

DEVICES = '/home/pi/MQP/degradation/devices'

_models_by_mac = {}

def get_function_instructions():
    mapping = {}
    for grouping in argv[1:]:
        element_list = grouping.split(',')
        mac_address = element_list[0]
        action_tuples = list(zip(element_list[1::2], element_list[2::2]))
        mapping[mac_address] = {}
        for function, instruction in action_tuples:
            mapping[mac_address][function] = instruction
    return mapping

def get_training_model_paths():
    paths = []
    for path in Path(DEVICES).rglob('training.model'):
        paths.append(str(path))
    return paths
    
def map_macs_to_models():
    print('Mapping MAC addresses to random forest models')
    global _models_by_mac
    cl_instructions = get_function_instructions()
    for training_model_path in get_training_model_paths():
        model, model_accuracy, mac_address, feature = load(open(training_model_path, 'rb'))
        if mac_address not in _models_by_mac:
            _models_by_mac[mac_address] = {}
        if mac_address in cl_instructions and feature in cl_instructions[mac_address]:
            instruction = cl_instructions[mac_address][feature]
            _models_by_mac[mac_address][feature] = (model, model_accuracy, instruction)
        else:
            _models_by_mac[mac_address][feature] = (model, model_accuracy, None)
    
def get_list_of_secured_macs():
    global _models_by_mac
    return [mac for mac in _models_by_mac]
    
def enact_iptables_filter(mac_address, queue_no):
    print('Enacting iptables filter: %s traffic to queue no. %d' % (mac_address, queue_no))
    c = 'sudo iptables -t raw -I PREROUTING -m mac --mac-source %s -j NFQUEUE --queue-num %d' % (mac_address, queue_no)
    print(c)
    check_output(c.split(' '))
    
def list_iptables_filters():
    c = 'sudo iptables -t raw -L'
    print(c)
    print(check_output(c.split(' ')).decode('ascii'))
    
def flush_iptables_filters():
    print('Flushing iptables filters and exposing all devices to rampant danger')
    c = 'sudo iptables -t raw -F'
    print(c)
    check_output(c.split())

def get_src_mac(nfq_packet):
    mac_string = b2a_hex(nfq_packet.get_hw()).decode('ascii')[0:12].upper()
    return ':'.join(mac_string[i:i+2] for i in range (0, len(mac_string), 2))
    
def make_prediction_against(model, list):
    return model.predict([list])
  
def get_most_confident_positive_ids(list_of_tuples):
    # list of (accuracy, instruction, prediction, label)
    positive_ids = [(a, i, p, l) for a, i, p, l in list_of_tuples if p == True]
    sorted_list = sorted(positive_ids, key=lambda x: x[0])
    #print(list_of_tuples)
    #print(sorted_list)
    #print()
    return sorted_list

def handle_nfq_packet(raw_packet):
    global _models_by_mac
    start_time = time()
    mac_address = get_src_mac(raw_packet)
    packet = IP(raw_packet.get_payload())
    feature_vector_list = get_packet_feature_list(packet)
    size_w_ether_frame = feature_vector_list[0] + 16 # missing ether frame
    feature_vector_list[0] = size_w_ether_frame
    function_meta_list = []
    for function_label in _models_by_mac[mac_address]:
        function_model = _models_by_mac[mac_address][function_label][0]
        function_model_accuracy = _models_by_mac[mac_address][function_label][1]
        function_instruction = _models_by_mac[mac_address][function_label][2]
        prediction = make_prediction_against(function_model, feature_vector_list)
        t = (function_model_accuracy, function_instruction, prediction, function_label)
        function_meta_list.append(t)
    most_confident_positive_ids_list = get_most_confident_positive_ids(function_meta_list)
    if len(most_confident_positive_ids_list) == 0:
        # drop the packet, not sure what it is
        raw_packet.accept()
        print('%s: unidentifiable packet, accepting (%f s)' % (mac_address, time() - start_time))
        return
    # by this point, we have a positive ID, i.e. mapping from packet to function
    accuracy, instruction, _, label = most_confident_positive_ids_list[0]
    if instruction == 'accept':
        raw_packet.accept()
        print('%s: %s packet, %d pct. confidence, accepting (%.4f s)' % (
            mac_address, label, int(accuracy*100), time() - start_time))
        return
    elif instruction == 'drop':
        raw_packet.drop()
        print('%s: %s packet, %d pct. confidence, dropping (%.4f s)' % (
            mac_address, label, int(accuracy*100), time() - start_time))
    else:
        raw_packet.accept()
        print('%s: %s packet, %d pct. confidence, ignoring (%.4f s)' % (
            mac_address, label, int(accuracy*100), time() - start_time))

def main():
    global _models_by_mac
    flush_iptables_filters()
    map_macs_to_models()
    nfqueue = NetfilterQueue()
    for i, mac_address in enumerate(get_list_of_secured_macs()):
        enact_iptables_filter(mac_address, i)
        nfqueue.bind(i, handle_nfq_packet)
    try:
        print('Actively filtering...')
        nfqueue.run()
    except KeyboardInterrupt:
        nfqueue.unbind()
        flush_iptables_filters()
        list_iptables_filters()
main()
