print('Loading libraries\n')

from pathlib import Path
from glob import glob
from pickle import load, dump
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn import metrics
import seaborn as sn
from scapy.all import *
from os.path import join
import create_training_set as ts

DEVICES = '/home/pi/MQP/degradation/devices'

feature_vector = {
    'size': [],
    'tcp_sport': [], 'tcp_dport': [], 'tcp_flags': [],
    'udp_sport': [], 'udp_dport': [],
    'is_function_associated': []
}

# Get list of devices with existing sets and/or models
def get_device_paths():
    return set(glob('%s/*' % DEVICES))

# Return set of paths to existing training sets
def get_training_set_paths(device_path):
    return set(glob('%s/*/training.set' % device_path))
    
# Merge on training set with a list of others
def get_merged_tuple_list(set_path, other_paths):
    print('Loading appropriate training sets')
    merged_tuple_list = []
    target_tuple_list = load(open(set_path, 'rb'))
    merged_tuple_list.extend(target_tuple_list)
    other_tuple_lists = [load(open(path, 'rb')) for path in other_paths]
    print('Marking appropriate packets as non-associated with this function')
    for other_tuple_list in other_tuple_lists:
        for index, other_tuple in enumerate(other_tuple_list):
            same_timestamp = other_tuple[0]
            same_packet = other_tuple[1]
            changed_flag = '' #!
            if set_path.split('/')[2] == 'baseline': changed_flag = not other_tuple[2] #!
            else: changed_flag = False #!
            #changed_flag = False
            other_tuple_list[index] = (same_timestamp, same_packet, changed_flag)
        merged_tuple_list.extend(other_tuple_list)
    return merged_tuple_list
    
def get_train_test_split(feature_vector_list):
    global feature_vector
    c = list(feature_vector.keys())
    # pd.set_option('display.max_rows', None)
    data_frame = pd.DataFrame(feature_vector_list, columns=c)
    # print(data_frame)
    x = data_frame[c[:-1]]
    y = data_frame[c[-1]]
    return train_test_split(x, y, test_size=0.25, random_state=0)

def get_random_forest_model(feature_vector_list):
    print('Getting random forest model')
    x_train, x_test, y_train, y_test = get_train_test_split(feature_vector_list)
    model = RandomForestClassifier(n_estimators=100)
    model.fit(x_train, y_train)
    return model
    
def get_rf_model_confidence(feature_vector_list, model):
    _, x_test, _, y_test = get_train_test_split(feature_vector_list)
    y_pred = model.predict(x_test)
    return metrics.accuracy_score(y_test, y_pred)
    
# Get feature packets from a packet
def get_packet_feature_list(packet):
    p = packet/IP()
    size = len(p)
    tcp_sport = 0
    tcp_dport = 0
    tcp_flags = 0
    if TCP in p:
        tcp_sport = p[TCP].sport
        tcp_dport = p[TCP].dport
        tcp_flags = int(hash(p[TCP].flags))
    udp_sport = 0
    udp_dport = 0
    if UDP in p:
        udp_sport = p[UDP].sport
        udp_dport = p[UDP].dport
    return [size, tcp_sport, tcp_dport, tcp_flags, udp_sport, udp_dport]
    
# Get a RF model from a list of merged tuples
def get_feature_vector_list(merged_tuple_list):
    print('Getting packet feature vector list')
    global feature_vector
    feature_vector_list = feature_vector
    for timestamp, packet, is_function_associated in merged_tuple_list:
        packet_feature_list = get_packet_feature_list(packet)
        feature_vector_list['size'].append(packet_feature_list[0])
        feature_vector_list['tcp_sport'].append(packet_feature_list[1])
        feature_vector_list['tcp_dport'].append(packet_feature_list[2])
        feature_vector_list['tcp_flags'].append(packet_feature_list[3])
        feature_vector_list['udp_sport'].append(packet_feature_list[4])
        feature_vector_list['udp_dport'].append(packet_feature_list[5])
        feature_vector_list['is_function_associated'].append(is_function_associated)
    return feature_vector_list
    
def clear_feature_vector():
    global feature_vector
    for key in feature_vector:
        feature_vector[key] = []

def main():
    for device_path in get_device_paths():
        print('Reviewing %s' % device_path)
        training_set_paths = get_training_set_paths(device_path)
        for training_set_path in training_set_paths:
            print('Reviewing %s' % training_set_path)
            clear_feature_vector() # vitally important, lol
            other_paths = training_set_paths - set([training_set_path])
            merged_tuple_list = get_merged_tuple_list(training_set_path, other_paths)
            feature_vector_list = get_feature_vector_list(merged_tuple_list)
            random_forest_model = get_random_forest_model(feature_vector_list)
            random_forest_model_confidence = get_rf_model_confidence(
                    feature_vector_list, random_forest_model)
            target_function = training_set_path.split('/')[-2]
            other_functions = [path.split('/')[-2] for path in other_paths]
            print('Classifying %s w/ noise from %s: %.2f\n' % 
                (target_function, str(other_functions), random_forest_model_confidence))
            # Create model...
            device_mac = training_set_path.split('/')[-3]
            pickle_dump = (random_forest_model, random_forest_model_confidence,
                    device_mac.upper().replace('.', ':'), target_function)
            pickle_path = join(ts.get_target_path(device_mac, target_function),
                    'training.model')
            print(pickle_dump, pickle_path, '\n')
            dump(pickle_dump, open(pickle_path, 'wb'))

if __name__ == '__main__':
    main()
